import unittest
import pydarm
import numpy as np


class TestComputeDigitalFilterResponse(unittest.TestCase):

    def setUp(self):
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_digital_filter_response = np.array(
            [-3513076500191.6504+3327250871150.924j,
             -164538201540.2522+209970221596.44916j,
             -3242877524.3140893-646390346.0757524j,
             3152750901.033689+60862553.6703552j,
             5968957387.406848+6106710482.173726j,
             17063001608.677845+14901038698.213606j,
             59173157596.065796-25364518140.6575j,
             -46407563930.73129+21858627480.823456j,
             -9021881.73649469-6742239.626013521j,
             -27770965.672582164+52593450.2554684j])

    def tearDown(self):
        del self.frequencies
        del self.known_digital_filter_response

    def test_compute_digital_filter_response(self):
        """ Test the DARM digital filter repsonse """
        D = pydarm.darm.DigitalModel('''
[metadata]
[interferometer]
[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules_in_use = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
''')
        test_response = D.compute_response(
            self.frequencies)
        for n in range(len(self.frequencies)):
            # Requires investigation why this delta tolerance has to be higher
            # on some systems in order to pass.
            # Can test using print() to check output and call test using
            # pytest -s test/darm_test.py
            self.assertAlmostEqual(
                np.abs(test_response[n]) /
                np.abs(self.known_digital_filter_response[n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_response[n], deg=True) -
                np.angle(self.known_digital_filter_response[n], deg=True), 0.0,
                places=5)


class TestComputeDarmOlg(unittest.TestCase):

    def setUp(self):
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_olg = np.array(
            [-346589.7289879694+67697.98179370671j,
             -1859.4410636292505+7152.898242047175j,
             40.79712693912431-10.496083469022562j,
             -3.200667505597764+0.6749176141222187j,
             -1.0964382128049217-0.6352470027806889j,
             -0.48849524504067693-0.12453097698086546j,
             0.07580387098963454+0.20189845660408226j,
             -0.010730361938738586+0.010017089450759862j,
             1.1367097388995271e-07-1.7573024415341907e-07j,
             -1.8569508267009735e-08+2.419265648650538e-08j])

    def tearDown(self):
        del self.frequencies
        del self.known_olg

    def test_compute_darm_olg(self):
        """ Test DARM open loop gain response """
        darm = pydarm.darm.DARMModel('''
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1638.001638001638, 1638.001638001638
omc_compensation_filter_file = test/H1OMC_1239468752.txt
omc_compensation_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_compensation_filter_modules_in_use = 4: 4
omc_compensation_filter_gain = 1, 1

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_meas_Z_UL = 129.7e3
tst_driver_meas_Z_LL = 90.74e3
tst_driver_meas_Z_UR = 93.52e3
tst_driver_meas_Z_LR = 131.5e3
tst_driver_meas_P_UL = 3.213e3, 31.5e3
tst_driver_meas_P_LL = 3.177e3, 26.7e3
tst_driver_meas_P_UR = 3.279e3, 26.6e3
tst_driver_meas_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules_in_use = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
''')
        test_olg = darm.compute_darm_olg(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(test_olg[n]) / np.abs(self.known_olg[n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_olg[n], deg=True) -
                np.angle(self.known_olg[n], deg=True), 0.0, places=5)


class TestComputeResponseFunction(unittest.TestCase):

    def setUp(self):
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_response = np.array(
            [2.0406181004597967-0.4013213677915312j,
             0.0011220990463462926-0.004469979302356652j,
             7.174381300234404e-06-1.4617952572908577e-06j,
             -6.479990333472527e-07+1.5653156901534748e-07j,
             -1.3662040618868035e-09-1.9861756619000937e-07j,
             1.6810351811864666e-07+2.0548446256238736e-08j,
             2.1237360469410807e-07+3.590194203666282e-07j,
             -7.07952406495893e-08+6.371681259820066e-07j,
             -1.4643321723110076e-06+3.0336488559992614e-07j,
             4.465169869308327e-06-1.5260050031500522e-06j])

    def tearDown(self):
        del self.frequencies
        del self.known_response

    def test_compute_darm_response_function(self):
        """ Test DARM closed loop response function """
        model_string = '''
[metadata]
[interferometer]
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1, 1

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_meas_Z_UL = 129.7e3
tst_driver_meas_Z_LL = 90.74e3
tst_driver_meas_Z_UR = 93.52e3
tst_driver_meas_Z_LR = 131.5e3
tst_driver_meas_P_UL = 3.213e3, 31.5e3
tst_driver_meas_P_LL = 3.177e3, 26.7e3
tst_driver_meas_P_UR = 3.279e3, 26.6e3
tst_driver_meas_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules_in_use = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
'''
        darm = pydarm.darm.DARMModel(model_string)
        test_response = darm.compute_response_function(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(test_response[n]) / np.abs(self.known_response[n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_response[n], deg=True) -
                np.angle(self.known_response[n], deg=True), 0.0, places=5)


class TestComputeEpicsRecords(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.array(
            [17.1, 15.6, 16.4, 17.6, 410.3, 1083.7, 7.93])
        self.known_epics = {
            'CAL-CS_TDEP_PCAL_LINE1_CORRECTION':
            0.003384106838174459+0.0004044180029064908j,
            'CAL-CS_TDEP_SUS_LINE3_REF_INVA_TST_RESPRATIO':
            1196925220371225.5-3841998698753872j,
            'CAL-CS_TDEP_SUS_LINE2_REF_INVA_PUM_RESPRATIO':
            -4.90204412692278e+17-1.9987201437993693e+17j,
            'CAL-CS_TDEP_SUS_LINE1_REF_INVA_UIM_RESPRATIO':
            2.643473241664109e+17-1.2695297228124422e+17j,
            'CAL-CS_TDEP_PCAL_LINE2_REF_C_NOCAVPOLE':
            3052033.7355621597-1030751.4694586383j,
            'CAL-CS_TDEP_PCAL_LINE2_REF_D':
            -243010006.4310854-1207741923.4087756j,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_TST':
            -3.6301159301930236e-19+5.17252131286287e-19j,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_PUM':
            -3.486603813748813e-21+2.854144501791321e-21j,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_UIM':
            9.669640696236252e-23+1.0084874176645285e-22j,
            'CAL-CS_TDEP_PCAL_LINE2_CORRECTION':
            5.919275925544666e-06+3.317567316630787e-07j,
            'CAL-CS_TDEP_PCAL_LINE4_REF_C_NOCAVPOLE':
            3219939.105380994-20248.45691682977j,
            'CAL-CS_TDEP_PCAL_LINE4_REF_D':
            -1334010103.6339288-1921974046.2409313j,
            'CAL-CS_TDEP_PCAL_LINE4_REF_A_TST':
            3.868595124883083e-17-7.793733735356113e-16j,
            'CAL-CS_TDEP_PCAL_LINE4_REF_A_PUM':
            -1.6867571291492549e-15+9.32181952713231e-16j,
            'CAL-CS_TDEP_PCAL_LINE4_REF_A_UIM':
            2.420118862241151e-16+3.3925102853622382e-16j,
            'CAL-CS_TDEP_PCAL_LINE4_CORRECTION':
            0.015159268559718102+0.003900900009459194j,
            'CAL-CS_TDEP_PCAL_LINE1_REF_C_NOCAVPOLE':
            3219708.6791652623-43662.16666949552j,
            'CAL-CS_TDEP_PCAL_LINE1_REF_D':
            3152795532.787565+60933868.87924543j,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_TST':
            -1.1944749080760831e-16-2.1720205784065532e-16j,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_PUM':
            -2.1758869487866404e-16+2.620157641796169e-16j,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_UIM':
            4.078008711853952e-17+4.992745126138772e-18j,
            'CAL-CS_TDEP_PCAL_LINE3_CORRECTION':
            8.319363469674813e-07+1.1949591405958398e-07j,
            'CAL-CS_TDEP_PCAL_LINE1_PCAL_DEMOD_PHASE': -0.37573242187500006,
            'CAL-CS_TDEP_PCAL_LINE2_PCAL_DEMOD_PHASE': -9.015380859375002,
            'CAL-CS_TDEP_PCAL_LINE3_PCAL_DEMOD_PHASE': -23.811767578125,
            'CAL-CS_TDEP_PCAL_LINE4_PCAL_DEMOD_PHASE': -0.17424316406249998,
            'CAL-CS_TDEP_SUS_LINE1_SUS_DEMOD_PHASE': -0.3427734375,
            'CAL-CS_TDEP_SUS_LINE2_SUS_DEMOD_PHASE': -0.36035156249999994,
            'CAL-CS_TDEP_SUS_LINE3_SUS_DEMOD_PHASE': -0.38671875}

    def tearDown(self):
        del self.frequencies
        del self.known_epics

    def test_compute_epics_records(self):
        model_string = '''
[metadata]
[interferometer]
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1, 1

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_meas_Z_UL = 129.7e3
tst_driver_meas_Z_LL = 90.74e3
tst_driver_meas_Z_UR = 93.52e3
tst_driver_meas_Z_LR = 131.5e3
tst_driver_meas_P_UL = 3.213e3, 31.5e3
tst_driver_meas_P_LL = 3.177e3, 26.7e3
tst_driver_meas_P_UR = 3.279e3, 26.6e3
tst_driver_meas_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[pcal]
pcal_dewhiten = 1.0, 1.0
ref_pcal_2_darm_act_sign = -1
analog_anti_aliasing_file = test/H1aa.mat
anti_aliasing_rate_string = 16k
anti_aliasing_method = biquad

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules_in_use = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
'''
        darm = pydarm.darm.DARMModel(model_string)
        test_epics = darm.compute_epics_records(
            self.frequencies[0], self.frequencies[1], self.frequencies[2],
            self.frequencies[3], self.frequencies[4], self.frequencies[5],
            self.frequencies[6], arm='x')
        for idx, (key, val) in enumerate(self.known_epics.items()):
            self.assertAlmostEqual(
                np.abs(test_epics[key]/self.known_epics[key]), 1.0, places=6)
            self.assertAlmostEqual(
                np.angle(test_epics[key], deg=True) -
                np.angle(self.known_epics[key], deg=True), 0.0, places=5)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy import signal, io

from .utils import digital_delay_filter
from .utils import compute_digital_filter_response, freqrespZPK
from .model import Model


class ActuationModel(Model):
    """
    An arm actuation model object

    This is a class to set up the model for the actuation function from a
    configuration file with all the information about where the data is stored

    Parameters
    ----------

    Returns
    -------


    """

    def __init__(self, config, measurement):
        super().__init__(config, measurement=measurement)

        # TODO need some way of validating the configuration file is
        # appropriate

    def analog_driver_response(self, frequencies):
        """
        The transfer function of the analog driver electronics. By default,
        the output will be a dictionary for UIM, PUM, and TST with
        dictionaries for each containing the quadrants UL, LL, UR, LR.
        Transfer functions are unity with zero phase unless the params
        file/string contains a <uim|pum|tst>_driver_meas_<z|p>_<ul|ll|ur|lr>
        for the particular quadrant and values

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        out : dict, `complex128`, array-like
            transfer function response of the uncompensated driver electronics
        """

        out = {'UIM': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')},
               'PUM': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')},
               'TST': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')}}

        for i, stage in enumerate(out.keys()):
            for j, quadrant in enumerate(out[stage].keys()):
                if (hasattr(self, f'{stage.lower()}_driver_meas_z_{quadrant.lower()}') and
                        hasattr(self, f'{stage.lower()}_driver_meas_p_{quadrant.lower()}')):
                    zeros = np.asarray(
                        getattr(self, f'{stage.lower()}_driver_meas_z_{quadrant.lower()}'))
                    poles = np.asarray(
                        getattr(self, f'{stage.lower()}_driver_meas_p_{quadrant.lower()}'))
                    gain = np.prod(2.0*np.pi*poles) / np.prod(2.0*np.pi*zeros)
                    model = signal.ZerosPolesGain(
                        -2.0*np.pi*zeros,
                        -2.0*np.pi*poles,
                        gain)
                    tf = signal.freqresp(model, 2.0*np.pi*frequencies)[1]
                    out[stage][quadrant] = tf

        return out

    def uim_dc_gain_Apct(self):
        """
        This computes the UIM DC gain in units of amps / count
        """

        return 4 * 0.25 * self.dac_gain * self.uim_driver_dc_trans_apv

    def pum_dc_gain_Apct(self):
        """
        This computes the PUM DC gain in units of amps / count
        """

        return 4 * 0.25 * self.dac_gain * self.pum_driver_dc_trans_apv * \
            self.pum_coil_outf_signflip

    def tst_dc_gain_V2pct(self):
        """
        This computes the TST DC gain in units of volts**2 / count
        """

        return 4 * 0.25 * self.dac_gain * self.tst_driver_dc_gain_vpv_lv * \
            2 * self.tst_driver_dc_gain_vpv_hv * \
            self.actuation_esd_bias_voltage

    def uim_dc_gain_Npct(self):
        """
        This computes the UIM DC gain in units of Newtons / count
        """

        return self.uim_dc_gain_Apct() * self.uim_npa

    def pum_dc_gain_Npct(self):
        """
        This computes the PUM DC gain in units of Newtons / count
        """

        return self.pum_dc_gain_Apct() * self.pum_npa

    def tst_dc_gain_Npct(self):
        """
        This computes the TST DC gain in units of Newtons / count
        """

        return self.tst_dc_gain_V2pct() * self.tst_npv2

    def sus_digital_filters_response(self, frequencies):
        """
        The transfer function of the SUS digital filters

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_digital_filter_response : `complex128`, array-like
            transfer function response of the digital filters in the UIM path
        pum_digital_filter_response : `complex128`, array-like
            transfer function response of the digital filters in the PUM path
        tst_digital_filter_response : `complex128`, array-like
            transfer function response of the digital filters in the TST path
        """

        # Start with TST stage filters
        tst_isc_inf_filter_response, pfilt = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.tst_isc_inf_bank, self.tst_isc_inf_modules[0],
                self.tst_isc_inf_gain, frequencies)
        tst_lock_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.tst_lock_bank, self.tst_lock_modules[0],
                self.tst_lock_gain, frequencies, pfilt)[0]
        tst_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.tst_drive_align_bank, self.tst_drive_align_modules[0],
                self.tst_drive_align_gain, frequencies, pfilt)[0]
        tst_digital_filter_response = tst_isc_inf_filter_response * \
            tst_lock_filter_response * tst_drive_align_filter_response

        # The PUM stage feedback is to the same SUS as TST stage, so
        # we can reuse the pfilt variable from above
        pum_lock_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.pum_lock_bank, self.pum_lock_modules[0],
                self.pum_lock_gain, frequencies, pfilt)[0]
        pum_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.pum_drive_align_bank, self.pum_drive_align_modules[0],
                self.pum_drive_align_gain, frequencies, pfilt)[0]
        pum_digital_filter_response = tst_isc_inf_filter_response * \
            tst_lock_filter_response * pum_lock_filter_response * \
            pum_drive_align_filter_response

        # The UIM stage feedback is to the same SUS as PUM stage, so
        # we can reuse the pfilt variable from above
        uim_lock_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.uim_lock_bank, self.uim_lock_modules[0],
                self.uim_lock_gain, frequencies, pfilt)[0]
        uim_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.uim_drive_align_bank, self.uim_drive_align_modules[0],
                self.uim_drive_align_gain, frequencies, pfilt)[0]
        uim_digital_filter_response = tst_isc_inf_filter_response * \
            tst_lock_filter_response * pum_lock_filter_response * \
            uim_lock_filter_response * uim_drive_align_filter_response

        return uim_digital_filter_response, \
            pum_digital_filter_response, \
            tst_digital_filter_response

    def sus_digital_compensation_response(self, frequencies):
        """
        The transfer function of the SUS compensation filters for driver
        electronics. By default, the output will be a dictionary for UIM,
        PUM, and TST with dictionaries for each containing the quadrants
        UL, LL, UR, LR. Transfer functions are unity with zero phase unless
        the params file/string contains:
        <uim|pum|tst>_compensation_filter_bank_<ul|ll|ur|lr> with a name;
        <uim|pum|tst>_compensation_filter_modules_in_use_<ul|ll|ur|lr> with
        list of modules turned on; and
        <uim|pum|tst>_compensation_filter_gain_<ul|ll|ur|lr>

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        out : dict, `complex128`, array-like
            transfer function response of the digital compensation filters in
            each path and quadrant
        """

        out = {'UIM': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')},
               'PUM': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')},
               'TST': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')}}

        pfilt = None
        for i, stage in enumerate(out.keys()):
            for j, quadrant in enumerate(out[stage].keys()):
                if hasattr(self, f'{stage.lower()}_compensation_filter_bank_{quadrant.lower()}'):
                    bank = getattr(self,
                                   f'{stage.lower()}_compensation_filter_bank_{quadrant.lower()}')
                if hasattr(
                        self,
                        f'{stage.lower()}_compensation_filter_modules_in_use_{quadrant.lower()}'):
                    modules = getattr(
                        self,
                        f'{stage.lower()}_compensation_filter_modules_in_use_{quadrant.lower()}')
                if hasattr(self, f'{stage.lower()}_compensation_filter_gain_{quadrant.lower()}'):
                    gain = getattr(
                        self, f'{stage.lower()}_compensation_filter_gain_{quadrant.lower()}')

                if ('bank' in locals() and len(bank) > 0 and
                        'modules' in locals() and 'gain' in locals()):
                    tf, pfilt = compute_digital_filter_response(
                        self.dpath(self.sus_filter_file),
                        bank, modules, gain, frequencies, pfilt)

                    out[stage][quadrant] *= tf
                    del bank, gain, modules

        return out

    def combine_sus_quadrants(self, frequencies):
        """
        Combine the digital and analog paths for the SUS quadrants on a
        per-stage and per-path basis. Meaning
        (0.25 x UIM UL digital x UIM UL analog +
        0.25 x UIM LL digital x UIM LL analog + ...) and so on for each
        stage

        Parameters
        ----------
        sus_dict_digital : dict, `complex`, array-like
            dictionary of digital compensation transfer function values
        sus_dict_analog : dict, `complex`, array-like
            dictionary of analog driver transfer function values

        Returns
        -------
        uim : dict, `complex128`, array-like
            transfer function response of the digital compensation filters in
            each path and quadrant for the UIM
        pum : dict, `complex128`, array-like
            transfer function response of the digital compensation filters in
            each path and quadrant for the PUM
        tst : dict, `complex128`, array-like
            transfer function response of the digital compensation filters in
            each path and quadrant for the TST
        """

        sus_dict_digital = self.sus_digital_compensation_response(frequencies)
        sus_dict_analog = self.analog_driver_response(frequencies)

        uim = 0.25 * (
            sus_dict_digital['UIM']['UL'] * sus_dict_analog['UIM']['UL'] +
            sus_dict_digital['UIM']['LL'] * sus_dict_analog['UIM']['LL'] +
            sus_dict_digital['UIM']['UR'] * sus_dict_analog['UIM']['UR'] +
            sus_dict_digital['UIM']['LR'] * sus_dict_analog['UIM']['LR'])
        pum = 0.25 * (
            sus_dict_digital['PUM']['UL'] * sus_dict_analog['PUM']['UL'] +
            sus_dict_digital['PUM']['LL'] * sus_dict_analog['PUM']['LL'] +
            sus_dict_digital['PUM']['UR'] * sus_dict_analog['PUM']['UR'] +
            sus_dict_digital['PUM']['LR'] * sus_dict_analog['PUM']['LR'])
        tst = 0.25 * (
            sus_dict_digital['TST']['UL'] * sus_dict_analog['TST']['UL'] +
            sus_dict_digital['TST']['LL'] * sus_dict_analog['TST']['LL'] +
            sus_dict_digital['TST']['UR'] * sus_dict_analog['TST']['UR'] +
            sus_dict_digital['TST']['LR'] * sus_dict_analog['TST']['LR'])

        return uim, pum, tst

    @staticmethod
    def matlab_force2length_response(suspension_file, frequencies):
        """
        Load the ZPK output from the matlab exported values and output a
        frequency response. This function expects variables in a .mat file to
        be UIMz, UIMp, UIMk, PUMz, PUMp, PUMk, TSTz, TSTp, and TSTk

        Parameters
        ----------
        suspension_file: `str`
            path and filename to .mat file that has the ZPK arrays
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        UIM_F2L_freqresp : `complex128`, array-like
            transfer function response of the UIM stage
        PUM_F2L_freqresp : `complex128`, array-like
            transfer function response of the PUM stage
        TST_F2L_freqresp : `complex128`, array-like
            transfer function response of the TST stage
        """

        mat = io.loadmat(suspension_file)
        UIM_F2L = signal.ZerosPolesGain(mat['UIMz'][:, 0], mat['UIMp'][:, 0],
                                        mat['UIMk'][0, 0])
        PUM_F2L = signal.ZerosPolesGain(mat['PUMz'][:, 0], mat['PUMp'][:, 0],
                                        mat['PUMk'][0, 0])
        TST_F2L = signal.ZerosPolesGain(mat['TSTz'][:, 0], mat['TSTp'][:, 0],
                                        mat['TSTk'][0, 0])
        UIM_F2L_freqresp = freqrespZPK(UIM_F2L, 2.0*np.pi*frequencies)
        PUM_F2L_freqresp = freqrespZPK(PUM_F2L, 2.0*np.pi*frequencies)
        TST_F2L_freqresp = freqrespZPK(TST_F2L, 2.0*np.pi*frequencies)

        return UIM_F2L_freqresp, PUM_F2L_freqresp, TST_F2L_freqresp

    def digital_out_to_displacement(self, frequencies):
        """
        This computes the transfer function from the output of the drive align
        filter bank of each stage to displacement of the test mass

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_response : `complex128`, array-like
            transfer function response of the UIM stage
        pum_response : `complex128`, array-like
            transfer function response of the PUM stage
        tst_response : `complex128`, array-like
            transfer function response of the TST stage
        """

        # SUS to IOP and IOP to analog digital time delays (see G1601472)
        sus_to_iop_delay_response = signal.dfreqresp(
            digital_delay_filter(1, 16384), 2.0*np.pi*frequencies/16384)[1]
        iop_to_analog_delay_response = signal.dfreqresp(
            digital_delay_filter(4, 65536), 2.0*np.pi*frequencies/65536)[1]

        # digital anti-imaging filter
        digital_ai_filter_response = \
            self.digital_aa_or_ai_filter_response(frequencies)

        # analog anti-imaging filter response
        analog_ai_filter_response = \
            self.analog_aa_or_ai_filter_response(frequencies)

        # Combine analog driver electronics for each quadrant with the
        # digital compensation for each quadrant, and then take the mean
        # of the four quadrants for each stage
        uim_tf, pum_tf, tst_tf = self.combine_sus_quadrants(frequencies)

        # Analog force to length response for each stage from Matlab
        sus_file = self.dpath(self.suspension_file)
        [uim_f2l_response,
         pum_f2l_response,
         tst_f2l_response] = self.matlab_force2length_response(sus_file,
                                                               frequencies)

        # Unknown overall time delay
        unknown_actuation_delay = np.exp(-2.0*np.pi*1j *
                                         self.unknown_actuation_delay *
                                         frequencies)

        # Independent time delays
        if hasattr(self, 'uim_delay') and self.uim_delay != '':
            uim_delay_tf = np.exp(-2.0*np.pi*1j*self.uim_delay*frequencies)
        else:
            uim_delay_tf = 1
        if hasattr(self, 'pum_delay') and self.pum_delay != '':
            pum_delay_tf = np.exp(-2.0*np.pi*1j*self.pum_delay*frequencies)
        else:
            pum_delay_tf = 1
        if hasattr(self, 'tst_delay') and self.tst_delay != '':
            tst_delay_tf = np.exp(-2.0*np.pi*1j*self.tst_delay*frequencies)
        else:
            tst_delay_tf = 1

        # Everything together now!
        tst_response = (tst_tf *
                        sus_to_iop_delay_response *
                        iop_to_analog_delay_response *
                        digital_ai_filter_response *
                        analog_ai_filter_response *
                        self.tst_dc_gain_Npct() *
                        tst_f2l_response *
                        unknown_actuation_delay *
                        tst_delay_tf)
        pum_response = (pum_tf *
                        sus_to_iop_delay_response *
                        iop_to_analog_delay_response *
                        digital_ai_filter_response *
                        analog_ai_filter_response *
                        self.pum_dc_gain_Npct() *
                        pum_f2l_response *
                        unknown_actuation_delay *
                        pum_delay_tf)
        uim_response = (uim_tf *
                        sus_to_iop_delay_response *
                        iop_to_analog_delay_response *
                        digital_ai_filter_response *
                        analog_ai_filter_response *
                        self.uim_dc_gain_Npct() *
                        uim_f2l_response *
                        unknown_actuation_delay *
                        uim_delay_tf)

        return uim_response, pum_response, tst_response

    def drivealign_to_displacement(self, frequencies):
        """
        This computes the transfer function from the input of the drive align
        filter bank of each stage to displacement of the test mass.

        This is particularly useful for computing EPICS records

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_response : `complex128`, array-like
            transfer function response of the UIM stage
        pum_response : `complex128`, array-like
            transfer function response of the PUM stage
        tst_response : `complex128`, array-like
            transfer function response of the TST stage
        """

        # Digital out to displacement of the TST stage by each of the
        # SUS stages
        [uim_response,
         pum_response,
         tst_response] = self.digital_out_to_displacement(frequencies)

        # Need to add in the DRIVEALIGN bank for each stage
        # Like before, we can reuse the pfilt variable if the preceeding
        # stage feedback goes to the same SUS
        tst_drive_align_filter_response, pfilt = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.tst_drive_align_bank, self.tst_drive_align_modules[0],
                self.tst_drive_align_gain, frequencies)
        pum_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.pum_drive_align_bank, self.pum_drive_align_modules[0],
                self.pum_drive_align_gain, frequencies, pfilt)[0]
        uim_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.uim_drive_align_bank, self.uim_drive_align_modules[0],
                self.uim_drive_align_gain, frequencies)[0]

        uim_response *= uim_drive_align_filter_response
        pum_response *= pum_drive_align_filter_response
        tst_response *= tst_drive_align_filter_response

        return uim_response, pum_response, tst_response

    def known_actuation_terms_for_measurement(self, frequencies):
        """
        This computes the transfer function from the input of the drive align
        filter bank of each stage to displacement of the test mass, everything
        but N/(drive unit); either amps or volts**2. This also includes the
        DARM feedback sign of the arm the measurement is taken on since we
        read out the measurment at DARM

        This is used for dividing out everything from measurements to establish
        the actuator gain in units of N/(drive unit)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_response : `complex128`, array-like
            transfer function response of the UIM stage, everything but N/ct
        pum_response : `complex128`, array-like
            transfer function response of the PUM stage, everything but N/ct
        tst_response : `complex128`, array-like
            transfer function response of the TST stage, everything but N/ct
        """

        [uim_response,
         pum_response,
         tst_response] = self.drivealign_to_displacement(frequencies)

        uim_response *= self.darm_feedback_sign * \
            (self.uim_dc_gain_Apct()/self.uim_dc_gain_Npct())
        pum_response *= self.darm_feedback_sign * \
            (self.pum_dc_gain_Apct()/self.pum_dc_gain_Npct())
        tst_response *= self.darm_feedback_sign * \
            (self.tst_dc_gain_V2pct()/self.tst_dc_gain_Npct())

        return uim_response, pum_response, tst_response

    def compute_actuation_single_stage(self, frequencies, stage='TST'):
        """
        Compute the actuation function transfer function for a single stage
        (see G1501518). This transfer function is from DARM_CTRL to meters
        sensed by the IFO. Note that the sign of the DARM_ERR signal is
        dependent upon which arm is under control.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the actuation function
        """

        # Digital output to TST displacement
        # Note that no feedback or output matrix values are applied here
        [uim_response,
         pum_response,
         tst_response] = self.digital_out_to_displacement(frequencies)
        if stage == 'UIM' or stage == 'L1':
            sus_response = uim_response
        elif stage == 'PUM' or stage == 'L2':
            sus_response = pum_response
        elif stage == 'TST' or stage == 'L3':
            sus_response = tst_response
        else:
            raise ValueError('stage must be UIM, PUM, TST, L1, L2, or L3')

        # All SUS digital filters
        [uim_digital_filter_response,
         pum_digital_filter_response,
         tst_digital_filter_response] = self.sus_digital_filters_response(
             frequencies)
        if stage == 'UIM' or stage == 'L1':
            digital_filter_response = uim_digital_filter_response
        elif stage == 'PUM' or stage == 'L2':
            digital_filter_response = pum_digital_filter_response
        else:
            digital_filter_response = tst_digital_filter_response

        # OMC to SUS digital time delay (see G1601472)
        # Note that SUS to IOP, IOP to analog, and unknown delays already
        # applied
        omc_to_sus_delay_response = signal.dfreqresp(
            digital_delay_filter(1, 16384), 2.0*np.pi*frequencies/16384)[1]

        sus_response *= digital_filter_response*omc_to_sus_delay_response

        # The response on DARM will depend on which arm is being actuated
        # by the DARM feedback
        sus_response *= self.darm_feedback_sign

        # Everything *above* here is modelling the physical actuator that we
        # have. However, the calibration group's convention is that A in their
        # loop drawings is minus the physical actuator (which allows the
        # explicit writing of the minus sign in the loop drawing to indicate
        # negative feedback). Ideally we'd just put
        # A = -1 * (TOP + UIM + PUM + TST),
        # but the actuation products below (which are used for other things
        # downstream) are expecting to receive minus of each stage of the
        # physical actuator. So, we multiply every stage by -1, and then sum
        # them up, and the result will be that A is negative of the physical
        # actuator. JCD_26Mar2019
        # See more details in T1800456.
        sus_response *= -1

        return sus_response


class DARMActuationModel(Model):
    """
    An DARM actuation model object

    This is a class to set up the model for the DARM loop from a
    configuration file with all the information about where the data is stored

    Parameters
    ----------

    Returns
    -------


    """

    def __init__(self, config):
        super().__init__(config, measurement='actuation')
        self.xarm = self.yarm = None
        if 'actuation_x_arm' in self._config:
            self.xarm = ActuationModel(config, measurement='actuation_x_arm')
        if 'actuation_y_arm' in self._config:
            self.yarm = ActuationModel(config, measurement='actuation_y_arm')

    def darm_output_matrix_values(self):
        """
        Turns the output matrix values into an array for the QUAD SUS
        """

        output_matrix = np.zeros((2, 4))
        for n in range(len(output_matrix[:, 0])):
            for m in range(len(output_matrix[0, :])):
                if n == 0 and self.darm_feedback_x[m] == 'ON':
                    output_matrix[n, m] = self.darm_output_matrix[0]
                elif n == 1 and self.darm_feedback_y[m] == 'ON':
                    output_matrix[n, m] = self.darm_output_matrix[1]

        return output_matrix

    def compute_actuation(self, frequencies, syserr_dict=None):
        """
        Compute the entire actuation function transfer function (see G1501518).
        This transfer function is from DARM_CTRL to meters sensed by the IFO.
        Note that the sign of the DARM_ERR signal is dependent upon which arm
        is under control.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        actuation_syserr_dict : `dict`, optional
            dict of multiplicative values, ex.:
            {'xarm': {'UIM': `complex`, array-like}}

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the actuation function
        """

        output_matrix = self.darm_output_matrix_values()

        darm_actuation = np.zeros(len(frequencies), dtype='complex128')
        for arm in range(len(output_matrix[:, 0])):
            for stage in range(len(output_matrix[0, :])):
                if output_matrix[arm, stage] != 0.0:
                    if stage == 0:
                        stage_label = 'TOP'
                    elif stage == 1:
                        stage_label = 'UIM'
                    elif stage == 2:
                        stage_label = 'PUM'
                    else:
                        stage_label = 'TST'
                    if arm == 0:
                        if self.xarm is None:
                            raise ValueError(
                                'Must provide an xarm ActuationModel object')
                        this_arm = self.xarm
                    else:
                        if self.yarm is None:
                            raise ValueError(
                                'Must provide a yarm ActuationModel object')
                        this_arm = self.yarm
                    single_stage = this_arm.compute_actuation_single_stage(
                        frequencies, stage=stage_label)

                    if (syserr_dict is not None and
                       ['xarm', 'yarm'][arm] in syserr_dict and
                       stage_label in syserr_dict[['xarm', 'yarm'][arm]]):
                        single_stage *= \
                            syserr_dict[['xarm', 'yarm'][arm]][stage_label]

                    darm_actuation += output_matrix[arm, stage] * single_stage

        return darm_actuation



# Copyright (C) Jameson Rollins (2021)
#               Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import os
import signal
import logging
import argparse

import numpy as np

from . import __version__
from .darm import DARMModel


logger = logging.getLogger('pydarm')
logger.setLevel(os.getenv('LOG_LEVEL', 'WARNING').upper())
formatter = logging.Formatter('%(message)s')
handler = logging.StreamHandler()
handler.setFormatter(formatter)
logger.addHandler(handler)


########################################

DEFAULT_FREQ = '0.01:5000:3000'

DATA_SAVE_FORMATS = ['.hdf5', '.h5']


def freq_from_spec(spec=None):
    """logarithmicly spaced frequency array, based on specification string

    Specification string should be of form 'START:[NPOINTS:]STOP'.  If
    `spec` is an array, then the array is returned as-is, and if it's
    None the DEFAULT_FREQ specification is used.

    """
    if isinstance(spec, np.ndarray):
        return spec
    elif spec is None:
        spec = DEFAULT_FREQ
    fspec = spec.split(':')
    if len(fspec) == 2:
        fspec = fspec[0], DEFAULT_FREQ.split(':')[1], fspec[1]
    return np.logspace(
        np.log10(float(fspec[0])),
        np.log10(float(fspec[2])),
        int(fspec[1]),
    )


########################################

parser = argparse.ArgumentParser(
    description="""pyDARM calibration modelling

""",
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
parser.add_argument(
    '--version', '-v', action='version', version=__version__)
subparsers = parser.add_subparsers(
    metavar='COMMAND',
)


def _subcommand(func, **kwargs):
    """helper function for adding a new subcommand to the parser

    This function helps in adding sub-commands to the argparse
    ArgumentParser.  The code to .

    Each sub-command should be encoded as a function, with
    arguments/options mapped to function keyword arguments.  The
    function docstring will be used as the sub-command description,
    with first line used as the help.  Additional keyword arguments
    are passed to the add_parser method.

    The subparser itself will be returned, and it's add_argument()
    method should be used to add arguments and options to the
    sub-command.

    The main() function below handles the parsing of arguments, and
    executing the sub-command function with the the arguments/options
    passed in as keyword arguments.

    """
    name = func.__name__.split('_')[1]
    sp = subparsers.add_parser(
        name,
        help=func.__doc__.splitlines()[0],
        description=func.__doc__.strip(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        **kwargs
    )
    sp.set_defaults(func=func)
    return sp

##########


def cmd_fr(model, freq, plot=True, save=False):
    """calculate model frequency response and plot

    """
    import h5py
    from matplotlib import pyplot as plt

    from .plot import BodePlot

    out_data_files = set()
    out_plot_files = set()
    if save:
        plot = False
        out_files = set(save)
        for path in out_files:
            if os.path.splitext(path)[1] in DATA_SAVE_FORMATS:
                out_data_files.add(path)
        out_plot_files = out_files - out_data_files

    ##########

    if os.path.splitext(model)[1] in DATA_SAVE_FORMATS:
        with h5py.File(model) as f:
            title = f.attrs.get('title', '')
            freq = f['Frequency'][:]
            G = f['Data']['DARM'][:]
            R = f['Data']['Response'][:]
            C = f['Data']['Sensing'][:]
            A = f['Data']['Actuation'][:]
            D = f['Data']['Digital'][:]

    else:
        freq = freq_from_spec(freq)
        DARM = DARMModel(model)
        title = f'{DARM.name} {DARM.run} {DARM.model_date}'
        G = DARM.compute_darm_olg(freq)
        R = DARM.compute_response_function(freq)
        C = DARM.sensing.compute_sensing(freq)
        A = DARM.actuation.compute_actuation(freq)
        D = DARM.digital.compute_response(freq)

    if plot or out_plot_files:
        bp = BodePlot(title=title)
        bp.plot(
            freq, G,
            label='DARM',
        )
        bp.plot(
            freq, C,
            label='Sensing',
            linestyle='--',
        )
        bp.plot(
            freq, A,
            label='Actuation',
            linestyle='--',
        )
        bp.plot(
            freq, D,
            label='Digital',
            linestyle='--',
        )
        bp.plot(
            freq, R,
            label='Response',
        )
        bp.legend()

        if out_plot_files:
            for path in out_plot_files:
                logger.info(f'saving plot: {path}')
                try:
                    bp.save(path)
                except Exception as e:
                    parser.exit(2, f'Error saving plot: {e}.\n')
        else:
            plt.show()

    if out_data_files:
        for path in out_data_files:
            logger.info(f'saving frequency response data: {path}')
            with h5py.File(path, 'w') as f:
                f.attrs['title'] = title
                f.create_dataset('Frequency', data=freq)
                g = f.create_group('Data')
                g.create_dataset('DARM', data=G)
                g.create_dataset('Sensing', data=C)
                g.create_dataset('Actuation', data=A)
                g.create_dataset('Digital', data=D)
                g.create_dataset('Response', data=R)


sp = _subcommand(cmd_fr, aliases=['plot'])
sp.add_argument(
    'model', metavar='MODEL',
    help='model INI config file, or hdf5 frequency response data')
sp.add_argument(
    '--freq', '-f', metavar='FLO:[NPOINTS:]FHI',
    help=f'logarithmic frequency array specification in Hz [{DEFAULT_FREQ}]')
group = sp.add_mutually_exclusive_group()
group.add_argument(
    '--save', '-s', metavar='PATH', action='append',
    help=("save plot (.png/.pdf/.svg) or response data (.hdf5/.h5) to file"
          "(may be specified multiple times)"))
group.add_argument(
    '--no-plot', '-np', action='store_false', dest='plot',
    help="suppress plotting")


########################################


def main():
    args = parser.parse_args()
    logger.debug(args)
    if 'func' not in args:
        parser.print_help()
        parser.exit()
    kwargs = dict(args._get_kwargs())
    del kwargs['func']
    args.func(**kwargs)


if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    main()



# Copyright (C) Evan Goetz (2022)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy import signal
from .utils import load_foton_export_tf, digital_delay_filter
from .digital import daqdownsamplingfilters
from .model import Model
from .darm import DARMModel


class CALCSModel(DARMModel):

    def __init__(self, config, calcs=None):
        """
        Initialize a CALCSModel object

        Note that any string or path to file string in `calcs` will
        overwrite anything in the `config` parameter string or path to file
        """
        super().__init__(config)
        if 'calcs' in self._config:
            self.calcs = Model(config, measurement='calcs')
        if calcs is not None:
            self.calcs = Model(calcs, measurement='calcs')
        if not hasattr(self, 'calcs'):
            raise ValueError('No CALCS parameters have been defined')

    def optical_response_ratio(self, frequencies):
        """
        This computes (opt resp)_foton / (opt resp)

        It is a bit confusing because the FOTON filter is the inverse
        sensing function. So internally this computes
        [1/(opt resp)/FOTON] = (opt resp)_foton / (opt resp)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies for the optical response ratio

        Returns
        -------
        foton_inv_sensing_interp : `complex128`, array-like
            ratio of interpolated values
        """

        # We need the coupled cavity LTI object
        coupled_cavity = self.sensing.optical_response(
            self.sensing.coupled_cavity_pole_frequency,
            self.sensing.detuned_spring_frequency,
            self.sensing.detuned_spring_q,
            pro_spring=self.sensing.is_pro_spring)

        # load inverse sensing data from foton file export (1/SRC_D2N)
        foton_freq, foton_tf = load_foton_export_tf(
            self.dpath(self.calcs.foton_invsensing_tf))

        # Take ratio of true optical response / foton response
        foton_inv_sensing_warp = (
            (1.0 / signal.freqresp(coupled_cavity, 2.0*np.pi*foton_freq)[1]) /
            foton_tf)

        # interpolate to the requested frequencies
        foton_inv_sensing_interp = np.interp(frequencies,
                                             foton_freq,
                                             foton_inv_sensing_warp)

        return foton_inv_sensing_interp

    def digital_out_to_displacement(self, frequencies):
        """
        Output of CALCS DRIVEALIGN to approximate TST displacement

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies for the CALCS approximated DARM actuation

        Returns
        -------
        xarm_uim : `complex128`, array-like
            X-arm UIM actuation approximation
        xarm_pum : `complex128`, array-like
            X-arm PUM actuation approximation
        xarm_tst : `complex128`, array-like
            X-arm TST actuation approximation
        yarm_uim : `complex128`, array-like
            Y-arm UIM actuation approximation
        yarm_pum : `complex128`, array-like
            Y-arm PUM actuation approximation
        yarm_tst : `complex128`, array-like
            Y-arm TST actuation approximation
        """

        arms = ['xarm', 'yarm']
        stages = ['uim', 'pum', 'tst']
        for m, arm in enumerate(arms):
            for n, stage in enumerate(stages):
                tf_dig_interp = np.ones(len(frequencies), dtype='complex128')
                tf_ana_interp = np.ones(len(frequencies), dtype='complex128')
                if hasattr(self.calcs, f'{arm}_{stage}_coiloutf'):
                    val = getattr(self.calcs, f'{arm}_{stage}_coiloutf')
                    if val != '':
                        [f, tf_dig] = load_foton_export_tf(self.dpath(val))
                        tf_dig_interp = np.interp(frequencies, f, tf_dig)
                if hasattr(self.calcs, f'{arm}_{stage}_analog'):
                    val = getattr(self.calcs, f'{arm}_{stage}_analog')
                    if val != '':
                        [f, tf_ana] = load_foton_export_tf(self.dpath(val))
                        tf_ana_interp = np.interp(frequencies, f, tf_ana)

                if hasattr(self.calcs, f'{arm}_output_matrix'):
                    out_mtrx = getattr(self.calcs, f'{arm}_output_matrix')[n + 1]
                else:
                    out_mtrx = 0

                vars()[f'{arm}_{stage}'] = out_mtrx * tf_dig_interp * tf_ana_interp

        return (vars()['xarm_uim'],
                vars()['xarm_pum'],
                vars()['xarm_tst'],
                vars()['yarm_uim'],
                vars()['yarm_pum'],
                vars()['yarm_tst'])

    def gds_sensing_correction(self, frequencies):
        """
        Compute the correction to the CAL-CS output for GDS

        We need to divide out the front end model optical response part only
        (filter is called usually SRC-D2N) and be sure to include 1 16k clock
        cycle delay from the model jump OMC to CAL-CS. There is an optical
        gain factor in the front end model, but it is fine to leave this in
        place. The output is basically C_pydarm / C_foton, and GDS will need to
        either divide the DARM_ERR data by this function or invert the output
        of this function and multiply by the DARM_ERR data.

        This is C_res*delay*(opt resp)/(opt resp)_foton

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the GDS sensing correction

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the correction
        """

        # Residual sensing function transfer function
        # This is everything in sensing except the optical gain and response
        C_res = self.sensing.sensing_residual(frequencies)

        # we will need to apply a delay because when it is inverted in GDS
        # this will become an advance (the correct thing to do)
        one_clock_cycle_filter = digital_delay_filter(1, 2**14)
        one_clock_cycle_filter_response = \
            signal.dfreqresp(one_clock_cycle_filter,
                             2.0*np.pi*frequencies/2**14)[1]

        # this is the sensing model without optical plant and with one clock
        # cycle delay applied
        C_res_with_calcs_delay = C_res * one_clock_cycle_filter_response

        # get the optical response ratio between FOTON and pyDARM
        foton_inv_sensing_interp = self.optical_response_ratio(frequencies)

        # the final correction is the sensing model divided by the interpolated
        # foton filter
        correction = C_res_with_calcs_delay / foton_inv_sensing_interp

        return correction

    def gds_actuation_correction(self, frequencies, arm, stage, daqdownsample=True):
        """
        Compute the correction to the CAL-CS output for GDS. Note that this
        implicitly assumes that the front end digital filters in CALCS is the
        same as that in the SUS path! If this is not the case, then this
        method will not return a correct answer

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the GDS sensing correction

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the correction
        """

        # Get the analog actuation transfer function from the front end
        foton_file = getattr(self.calcs,
                             f'{arm.lower()}_{stage.lower()}_analog')
        freq_foton, tf_foton = load_foton_export_tf(self.dpath(foton_file))

        # Get the full actuation for the stage and the digital filters
        if arm == 'xarm':
            sus_act = self.actuation.xarm.compute_actuation_single_stage(
                freq_foton, stage=stage)
            uim_filt, pum_filt, tst_filt = (
                self.actuation.xarm.sus_digital_filters_response(freq_foton))
        elif arm == 'yarm':
            sus_act = self.actuation.xarm.compute_actuation_single_stage(
                freq_foton, stage=stage)
            uim_filt, pum_filt, tst_filt = (
                self.actuation.xarm.sus_digital_filters_response(freq_foton))
        else:
            raise ValueError('arm must be either xarm or yarm')

        # CALCS output matrix
        out_matrix = getattr(self.calcs, f'{arm.lower()}_output_matrix')

        # This division here assumes that the CALCS FOTON filters match exactly
        # the SUS FOTON filters
        if stage == 'UIM':
            sus_act /= uim_filt * out_matrix[1]
        elif stage == 'PUM':
            sus_act /= pum_filt * out_matrix[2]
        elif stage == 'TST':
            sus_act /= tst_filt * out_matrix[3]
        else:
            raise ValueError('stage must be UIM, PUM, or TST')

        sus_act /= tf_foton
        # Now sus_act is basically A_i residual

        # Interpolate to the desired frequencies
        sus_act = np.interp(frequencies, freq_foton, sus_act)

        # Here we want the 1 clock cycle delay transfer function
        omc_to_sus = signal.dfreqresp(digital_delay_filter(1, 16384),
                                      2.0*np.pi*frequencies/16384)[1]

        if daqdownsample:
            # DAQ downsampling filters are applied so we need to account
            # for this
            daqdownsampling = signal.dfreqresp(
                daqdownsamplingfilters(2**14, 2**12, 'biquad', 'v3'),
                2.0*np.pi*frequencies/2**14)[1]
        else:
            daqdownsampling = 1

        # This removes the DAQ downsampling filters and the OMC to CAL-CS delay
        residual_actuation_filter = (sus_act / (daqdownsampling * omc_to_sus))

        return residual_actuation_filter

    def calcs_darm_actuation(self, frequencies):
        """
        Compute the CALCS approximated DARM actuation. This method implicitly
        assumes that the CALCS DARM output matrix and actuation digital
        filtering matches the installed in-loop DARM ouput matrix and actuation
        digital filtering. If they do not agree, then this method may give
        incorrect results, so it's best to confirm agreement

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies for the CALCS approximated DARM actuation

        Returns
        -------
        calcs_actuation : `complex128`, array-like
            Residual phase after removing the simulated delay
        """

        # Assume the same as in-loop DARM output matrix
        output_matrix = self.actuation.darm_output_matrix_values()

        # Compute the CALCS approximated "digitial out to displacement" =
        # output of CALCS DRIVEALIGN to displacement.
        # Order is UIM_x, PUM_x, TST_x, UIM_y, PUM_y, and TST_y
        calcs_dig_out_to_disp = self.digital_out_to_displacement(frequencies)

        # Put calcs_dig_out_to_disp into same order as the output matrix
        calcs_dig_out_to_disp = np.array(
            [[np.zeros(len(calcs_dig_out_to_disp[0]), dtype='complex128'),
              calcs_dig_out_to_disp[0],
              calcs_dig_out_to_disp[1],
              calcs_dig_out_to_disp[2]],
             [np.zeros(len(calcs_dig_out_to_disp[3]), dtype='complex128'),
              calcs_dig_out_to_disp[3],
              calcs_dig_out_to_disp[4],
              calcs_dig_out_to_disp[5]]])

        # Start with zeros
        calcs_actuation = np.zeros(len(frequencies), dtype='complex128')

        # Loop over arms and stages
        for arm in range(len(output_matrix[:, 0])):
            if np.any(output_matrix[arm, :]):
                if arm == 0:
                    dig_filt = (
                        self.actuation.xarm.sus_digital_filters_response(
                            frequencies))
                elif arm == 1:
                    dig_filt = (
                        self.actuation.yarm.sus_digital_filters_response(
                            frequencies))
            for stage in range(len(output_matrix[0, :])):
                if output_matrix[arm, stage] != 0.0:
                    # dig_filt only has UIM, PUM, and TST
                    # calcs_dig_out_to_disp has same order as output_matrix
                    calcs_actuation += (output_matrix[arm, stage] *
                                        dig_filt[stage - 1] *
                                        calcs_dig_out_to_disp[arm, stage])

        return calcs_actuation

    def residual_sensing_sim_delay(self, frequencies, clock=False):
        """
        Fit a simulated delay based on the residual sensing in units
        of seconds delay or (if clock=True) in units of 16384 samples/sec,
        i.e., 1 16384 Hz clock delay = 6.1035e-5 s. Positive indicates
        a delay, negative indicates an advance

        Internally, the fit domain is over the range of 10 Hz to 2000 Hz,
        though if your frequency range is smaller it will fit over the
        smaller.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to model the phase delay
        clock : `boolean`, optional
            if clock is True, then the output delay_fit and residual will
            be computed and rounded to the nearest number of integer clock
            cycles (16384)

        Returns
        -------
        delay_fit : float
            Best fit to the data; either in units of seconds or rounded to
            the nearest integer clock cycle (if clock=True)
        residual : `float`, array-like
            Residual phase after removing the simulated delay
        """

        C_res = self.sensing.sensing_residual(frequencies)

        phase_delay = np.angle(C_res) / (2.0*np.pi*frequencies)

        poly = np.polynomial.Polynomial.fit(frequencies, phase_delay, 0,
                                            domain=[10, 2000])
        delay_fit = -poly.convert().coef[0]

        if clock is True:
            delay_fit = np.round(delay_fit * 16384) / 16384

        residual = np.angle(C_res) + delay_fit*(2.0*np.pi*frequencies)

        if clock is True:
            delay_fit *= 16384

        return delay_fit, residual

    def residual_actuation_sim_delay(self, frequencies, clock=False):
        """
        Fit a simulated delay based on the residual actuation in units
        of seconds delay or (if clock=True) in units of 16384 samples/sec,
        i.e., 1 16384 Hz clock delay = 6.1035e-5 s. Positive indicates
        a delay, negative indicates an advance

        Internally, the fit domain is over the range of 10 Hz to 2000 Hz,
        though if your frequency range is smaller it will fit over the
        smaller.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to model the phase delay
        clock : `boolean`, optional
            if clock is True, then the output delay_fit and residual will
            be computed and rounded to the nearest number of integer clock
            cycles (16384)

        Returns
        -------
        delay_fit : float
            Best fit to the data; either in units of seconds or rounded to
            the nearest integer clock cycle (if clock=True)
        residual : `float`, array-like
            Residual phase after removing the simulated delay
        """

        A_res = (self.actuation.compute_actuation(frequencies) /
                 self.calcs_darm_actuation(frequencies))

        phase_delay = np.angle(A_res) / (2.0*np.pi*frequencies)

        poly = np.polynomial.Polynomial.fit(frequencies, phase_delay, 0,
                                            domain=[10, 2000])
        delay_fit = -poly.convert().coef[0]

        if clock is True:
            delay_fit = np.round(delay_fit * 16384) / 16384

        residual = np.angle(A_res) + delay_fit*(2.0*np.pi*frequencies)

        if clock is True:
            delay_fit *= 16384

        return delay_fit, residual

    def calcs_dtt_calibration(self, frequencies, include_whitening=True,
                              strain_calib=False, save_to_file=None):
        """
        Compute the calibration transfer function for the main control room
        calibrated sensitivity curve. One can save this data to a file
        with the needed frequency, dB magnitude, and degrees phase columns.

        Details (see T1900169):
        CAL-DELTAL_EXTERNAL = whiten * [d_err/C_foton + d_ctrl*A_foton] * delay
        where whiten is a whitening filter and delay is the 1 16384 clock cycle
        delay between the OMC model and CALCS
        real DELTAL_EXTERNAL = [d_err/C_real + d_ctrl*A_real]
        C_real != C_foton and A_real != A_foton, so the correction factor
        (or calibration factor) is going to be the ratio of
        real DELTAL_EXTERNAL / CAL-DELTAL_EXTERNAL
        So the calibration scaling is a multiplicative factor given by:
        (1+G)/(W*delay*C_r)*(opt resp)_foton/(opt resp) *
          [1+G/(C_r*A_r)*(opt resp)_foton/(opt resp)*FE]
        where G = C*D*A (the open loop gain)
        C_r = C/[H*(opt resp)]
        A_r = A/A_foton
        FE = the clock cycle delay or Thiran filter delay installed
             in the front end to delay the actuation path
        (see eq. 38 of T1900169)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the CAL_DELTAL_EXTERNAL DTT
            calibration
        include_whitening : `bool`, optional
            if the whitening filter is on (default), then we'd like to remove
            its effect so by default this divides out the whitening filter
        strain_calib : `bool`, optional
            the output defaults to dL_true/dL_calcs, so that we can recalibrate
            CAL-DELTAL_EXTERNAL. If this option is True, then the output is
            h_true/dL_calcs
        save_to_file : `str`, optional
            Filename (ASCII) to save the data from this result. Note:
            the file columns are <frequency> <magnitude (dB)> <phase (deg)>

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the calibration
        """

        # This is C_r*delay*(opt resp)/(opt resp)_foton
        gds_corr = self.gds_sensing_correction(frequencies)

        # Whitening filter
        foton_freq, foton_tf = load_foton_export_tf(
            self.dpath(self.calcs.foton_deltal_whitening_tf))
        whitening_interp = np.interp(frequencies,
                                     foton_freq,
                                     foton_tf)
        if not include_whitening:
            whitening_interp = 1

        # Foton delay filter
        foton_freq, foton_tf = load_foton_export_tf(
            self.dpath(self.calcs.foton_delay_filter_tf))
        delay_interp = np.interp(frequencies, foton_freq, foton_tf)

        # Model loop gain CDA
        G = self.compute_darm_olg(frequencies)

        # optical response ratio
        foton_inv_sensing_interp = self.optical_response_ratio(frequencies)

        # Residual sensing is sensing / (gain * opt response)
        C_res = self.sensing.sensing_residual(frequencies)

        # Residual actuation is actuation / A_foton
        A_res = (self.actuation.compute_actuation(frequencies) /
                 self.calcs_darm_actuation(frequencies))

        # Old: calib = 1.0 / (whitening_interp * gds_corr)
        # New:
        calib = ((1.0 + G) / (whitening_interp * gds_corr) /
                 (1.0 + G*foton_inv_sensing_interp*delay_interp/(C_res*A_res)))

        if strain_calib:
            calib /= np.mean(np.array([self.sensing.x_arm_length,
                                       self.sensing.y_arm_length]))

        if save_to_file is not None:
            np.savetxt(save_to_file,
                       np.array([frequencies,
                                 20.0*np.log10(np.abs(calib)),
                                 np.angle(calib, deg=True)]).T,
                       fmt='%.7g')

        return calib



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy import signal
from gwpy.time import tconvert, from_gps

from .model import Model
from .sensing import SensingModel
from .actuation import DARMActuationModel
from .pcal import PcalModel
from .utils import compute_digital_filter_response
from .digital import daqdownsamplingfilters
from .plot import critique


class DigitalModel(Model):
    """DARM digital controller model object

    """
    def __init__(self, config):
        super().__init__(config, measurement='digital')

    def compute_response(self, frequencies):
        """
        Compute DARM digital controller frequency response

        Uses filter ZPK transfer function response from Foton file.

        Parameters
        ----------
        frequencies

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the digital SUS filter

        """

        response = np.ones(len(frequencies), dtype='complex128')
        for n in range(len(self.digital_filter_bank)):
            if n == 0:
                tf, pfilt = compute_digital_filter_response(
                    self.dpath(self.digital_filter_file),
                    self.digital_filter_bank[n],
                    self.digital_filter_modules_in_use[n],
                    self.digital_filter_gain[n], frequencies, pfilt=None)
            else:
                tf = compute_digital_filter_response(
                    self.dpath(self.digital_filter_file),
                    self.digital_filter_bank[n],
                    self.digital_filter_modules_in_use[n],
                    self.digital_filter_gain[n], frequencies, pfilt=pfilt)[0]
            response *= tf
        return response


class DARMModel(Model):
    """DARM model object

    This is a class to set up the model for the DARM loop from a
    configuration file with all the information about where the data is stored

    """
    def __init__(self, config, sensing=None, actuation=None, digital=None, pcal=None):
        super().__init__(config)
        self.sensing = sensing or SensingModel(config)
        self.actuation = actuation or DARMActuationModel(config)
        self.digital = digital or DigitalModel(config)
        self.pcal = pcal
        if not self.pcal and 'pcal' in self._config:
            self.pcal = PcalModel(config)

    def compute_darm_olg(self, frequencies):
        """
        Compute the entire DARM open loop transfer function (see G1501518)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the sensing function

        """
        C_response = self.sensing.compute_sensing(frequencies)
        A_response = self.actuation.compute_actuation(frequencies)
        D_response = self.digital.compute_response(frequencies)

        return C_response * A_response * D_response

    def compute_response_function(self, frequencies, sensing_syserr=None,
                                  actuation_syserr_dict=None):
        """
        Compute the entire DARM response transfer function (see G1501518)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        sensing_syserr : `complex`, array-like, optional
            multiplicative factor to include relative sensing systematic error
        actuation_syserr_dict : `dict`, optional
            dict of multiplicative values, ex.:
            {'xarm': {'UIM': `complex`, array-like}}

        Returns
        -------
        tf : `complex128`, array-like
            transfer function of the DARM closed-loop response
        """

        C_response = self.sensing.compute_sensing(frequencies)
        A_response = self.actuation.compute_actuation(frequencies,
                                                      actuation_syserr_dict)
        D_response = self.digital.compute_response(frequencies)

        if sensing_syserr is not None:
            C_response *= sensing_syserr

        return (1.0/C_response + D_response * A_response)

    def compute_epics_records(self, f_pcal1, f_uim, f_pum, f_tst, f_pcal2,
                              f_pcal3, f_pcal4, arm, save_to_file=None):
        """
        This is a LIGO specific function for generating front end calibration
        and GDS EPICS records. This function is very specific for O3, but we
        may wish to extend it or generalize it for O4 and/or other detectors.
        Currently it accepts 4 PCAL injections frequencies and 3 SUS actuator
        frequencies.

        See T1700106-v10 and G1601472 for additional details

        Parameters
        ----------
        f_pcal1 : `float`
            Pcal frequency line 1 (the frequency for the SUS actuators)
        f_uim : `float`
            UIM line frequency
        f_pum : `float`
            PUM line frequency
        f_tst : `float`
            TST line frequency
        f_pcal2 : `float`
            Pcal line frequency 2 (the mid-frequency ~400 Hz line)
        f_pcal3 : `float`
            Pcal line frequency 3 (the high-frequency ~1080 Hz line)
        f_pcal4 : `float`
            Pcal line frequency 4 (the low frequency ~8 Hz line)
        arm : `str`
            'x' or 'y' to indicate which arm is used for the SUS actuators
        save_to_file : `str`, optional
            Filename (ASCII) to save the values for EPICS from this reference model


        Returns
        -------
        out : dict
            Dictionary of EPICS records

        """

        freq = np.array(
            [f_pcal1, f_uim, f_pum, f_tst, f_pcal2, f_pcal3, f_pcal4])

        R = self.compute_response_function(freq)

        daqdownsampling = daqdownsamplingfilters(2**14, 2**9, 'biquad', 'v3')
        daqdownsampling_response = signal.dfreqresp(daqdownsampling,
                                                    2.0*np.pi*freq/2**14)[1]

        # Pcal corrections
        # See T1700106
        pcal_correction = self.pcal.compute_pcal_correction(freq)
        PCAL_LINE1_CORRECTION = pcal_correction[0] * \
            np.exp(-2*np.pi*1j*freq[0]/16384)   # EP0
        PCAL_LINE2_CORRECTION = pcal_correction[4] * \
            np.exp(-2*np.pi*1j*freq[4]/16384)   # EP9
        PCAL_LINE3_CORRECTION = pcal_correction[5] * \
            np.exp(-2*np.pi*1j*freq[5]/16384)   # EP24
        PCAL_LINE4_CORRECTION = pcal_correction[6] * \
            np.exp(-2*np.pi*1j*freq[6]/16384)   # EP15

        # The actuator lines are injected at the input to the DRIVEALIGN bank
        # Also get the complete actuation model
        if arm == 'x':
            this_arm = self.actuation.xarm
        elif arm == 'y':
            this_arm = self.actuation.yarm
        else:
            raise ValueError('Must provide value of x or y to argument arm')

        uim_actuation_epics, pum_actuation_epics, tst_actuation_epics = \
            this_arm.drivealign_to_displacement(freq)
        uim_actuation_epics *= this_arm.darm_feedback_sign
        pum_actuation_epics *= this_arm.darm_feedback_sign
        tst_actuation_epics *= this_arm.darm_feedback_sign
        uim = this_arm.compute_actuation_single_stage(freq, stage='UIM')
        pum = this_arm.compute_actuation_single_stage(freq, stage='PUM')
        tst = this_arm.compute_actuation_single_stage(freq, stage='TST')

        # Actuator EPICS
        # Need to divide out a DAQ downsampling filter from the reference A
        # values because channels sampled at 512 Hz have an AA filter applied
        # See T1700106
        SUS_LINE1_REF_INVA_UIM_RESPRATIO = \
            (1.0/(uim_actuation_epics[1]/daqdownsampling_response[1])) * \
            (1.0/R[0]) * R[1] * np.exp(2*np.pi*1j*freq[1]/16384)  # EP3
        SUS_LINE2_REF_INVA_PUM_RESPRATIO = \
            (1.0/(pum_actuation_epics[2]/daqdownsampling_response[2])) * \
            (1.0/R[0]) * R[2] * np.exp(2*np.pi*1j*freq[2]/16384)  # EP2
        SUS_LINE3_REF_INVA_TST_RESPRATIO = \
            (1.0/(tst_actuation_epics[3]/daqdownsampling_response[3])) * \
            (1.0/R[0]) * R[3] * np.exp(2*np.pi*1j*freq[3]/16384)  # EP1

        # Need to add a phase delay to the PCAL readback because the
        # PCAL channel acquires a phase delay from the model jump
        # to CAL-CS and the calculation expects no delay (as in GDS)
        PCAL_LINE1_PCAL_DEMOD_PHASE = \
            np.angle(np.exp(-2*np.pi*1j*freq[0]/16384), deg=True)
        PCAL_LINE2_PCAL_DEMOD_PHASE = \
            np.angle(np.exp(-2*np.pi*1j*freq[4]/16384), deg=True)
        PCAL_LINE3_PCAL_DEMOD_PHASE = \
            np.angle(np.exp(-2*np.pi*1j*freq[5]/16384), deg=True)
        PCAL_LINE4_PCAL_DEMOD_PHASE = \
            np.angle(np.exp(-2*np.pi*1j*freq[6]/16384), deg=True)

        # Need to add a phase delay to the SUS excitation because the
        # SUS channel acquires a phase delay from the model jump
        # to CAL-CS and the calculation expects no delay (as in GDS).
        # Note that if a replica SUS excitation is used in the CAL-CS model
        # then these should actually be set to zero.
        SUS_LINE1_SUS_DEMOD_PHASE = \
            np.angle(np.exp(-2*np.pi*1j*freq[1]/16384), deg=True)
        SUS_LINE2_SUS_DEMOD_PHASE = \
            np.angle(np.exp(-2*np.pi*1j*freq[2]/16384), deg=True)
        SUS_LINE3_SUS_DEMOD_PHASE = \
            np.angle(np.exp(-2*np.pi*1j*freq[3]/16384), deg=True)

        # Compute the sensing function without the optical response but need to
        # include the optical gain
        coupled_cavity = self.sensing.optical_response(
            self.sensing.coupled_cavity_pole_frequency,
            self.sensing.detuned_spring_frequency,
            self.sensing.detuned_spring_q,
            pro_spring=self.sensing.is_pro_spring)
        sensing_no_cavity_response = \
            self.sensing.compute_sensing(freq) / \
            signal.freqresp(coupled_cavity, 2.0*np.pi*freq)[1]
        PCAL_LINE2_REF_C_NOCAVPOLE = sensing_no_cavity_response[4]  # EP4
        PCAL_LINE4_REF_C_NOCAVPOLE = sensing_no_cavity_response[6]  # EP10
        PCAL_LINE1_REF_C_NOCAVPOLE = sensing_no_cavity_response[0]  # EP19

        # DARM digital filtering
        darm_digital_filter_response = \
            self.digital.compute_response(freq)
        PCAL_LINE2_REF_D = darm_digital_filter_response[4]  # EP5
        PCAL_LINE4_REF_D = darm_digital_filter_response[6]  # EP11
        PCAL_LINE1_REF_D = darm_digital_filter_response[0]  # EP20

        # Actuation
        PCAL_LINE2_REF_A_TST = tst[4]  # EP6
        PCAL_LINE2_REF_A_PUM = pum[4]  # EP7
        PCAL_LINE2_REF_A_UIM = uim[4]  # EP8
        PCAL_LINE4_REF_A_TST = tst[6]  # EP12
        PCAL_LINE4_REF_A_PUM = pum[6]  # EP13
        PCAL_LINE4_REF_A_UIM = uim[6]  # EP14
        PCAL_LINE1_REF_A_TST = tst[0]  # EP21
        PCAL_LINE1_REF_A_PUM = pum[0]  # EP22
        PCAL_LINE1_REF_A_UIM = uim[0]  # EP23

        if save_to_file is not None:
            save_txt_epcis_value = np.vstack(
                (PCAL_LINE1_CORRECTION.real,
                 PCAL_LINE1_CORRECTION.imag,
                 SUS_LINE3_REF_INVA_TST_RESPRATIO.real,
                 SUS_LINE3_REF_INVA_TST_RESPRATIO.imag,
                 SUS_LINE2_REF_INVA_PUM_RESPRATIO.real,
                 SUS_LINE2_REF_INVA_PUM_RESPRATIO.imag,
                 SUS_LINE1_REF_INVA_UIM_RESPRATIO.real,
                 SUS_LINE1_REF_INVA_UIM_RESPRATIO.imag,
                 PCAL_LINE2_REF_C_NOCAVPOLE.real,
                 PCAL_LINE2_REF_C_NOCAVPOLE.imag,
                 PCAL_LINE2_REF_D.real,
                 PCAL_LINE2_REF_D.imag,
                 PCAL_LINE2_REF_A_TST.real,
                 PCAL_LINE2_REF_A_TST.imag,
                 PCAL_LINE2_REF_A_PUM.real,
                 PCAL_LINE2_REF_A_PUM.imag,
                 PCAL_LINE2_REF_A_UIM.real,
                 PCAL_LINE2_REF_A_UIM.imag,
                 PCAL_LINE2_CORRECTION.real,
                 PCAL_LINE2_CORRECTION.imag,
                 PCAL_LINE4_REF_C_NOCAVPOLE.real,
                 PCAL_LINE4_REF_C_NOCAVPOLE.imag,
                 PCAL_LINE4_REF_D.real,
                 PCAL_LINE4_REF_D.imag,
                 PCAL_LINE4_REF_A_TST.real,
                 PCAL_LINE4_REF_A_TST.imag,
                 PCAL_LINE4_REF_A_PUM.real,
                 PCAL_LINE4_REF_A_PUM.imag,
                 PCAL_LINE4_REF_A_UIM.real,
                 PCAL_LINE4_REF_A_UIM.imag,
                 PCAL_LINE4_CORRECTION.real,
                 PCAL_LINE4_CORRECTION.imag,
                 PCAL_LINE1_REF_C_NOCAVPOLE.real,
                 PCAL_LINE1_REF_C_NOCAVPOLE.imag,
                 PCAL_LINE1_REF_D.real,
                 PCAL_LINE1_REF_D.imag,
                 PCAL_LINE1_REF_A_TST.real,
                 PCAL_LINE1_REF_A_TST.imag,
                 PCAL_LINE1_REF_A_PUM.real,
                 PCAL_LINE1_REF_A_PUM.imag,
                 PCAL_LINE1_REF_A_UIM.real,
                 PCAL_LINE1_REF_A_UIM.imag,
                 PCAL_LINE3_CORRECTION.real,
                 PCAL_LINE3_CORRECTION.imag,
                 PCAL_LINE1_PCAL_DEMOD_PHASE,
                 PCAL_LINE2_PCAL_DEMOD_PHASE,
                 PCAL_LINE3_PCAL_DEMOD_PHASE,
                 PCAL_LINE4_PCAL_DEMOD_PHASE,
                 SUS_LINE1_SUS_DEMOD_PHASE,
                 SUS_LINE2_SUS_DEMOD_PHASE,
                 SUS_LINE3_SUS_DEMOD_PHASE))

            save_txt_epics_channel_name = np.vstack(
                ('CAL-CS_TDEP_PCAL_LINE1_CORRECTION_REAL',
                 'CAL-CS_TDEP_PCAL_LINE1_CORRECTION_IMAG'
                 'CAL-CS_TDEP_SUS_LINE3_REF_INVA_TST_RESPRATIO_REAL',
                 'CAL-CS_TDEP_SUS_LINE3_REF_INVA_TST_RESPRATIO_IMAG',
                 'CAL-CS_TDEP_SUS_LINE2_REF_INVA_PUM_RESPRATIO_REAL',
                 'CAL-CS_TDEP_SUS_LINE2_REF_INVA_PUM_RESPRATIO_IMAG',
                 'CAL-CS_TDEP_SUS_LINE1_REF_INVA_UIM_RESPRATIO_REAL',
                 'CAL-CS_TDEP_SUS_LINE1_REF_INVA_UIM_RESPRATIO_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_C_NOCAVPOLE_REAL',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_C_NOCAVPOLE_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_D_REAL',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_D_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_A_TST_REAL',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_A_TST_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_A_PUM_REAL',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_A_PUM_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_A_UIM_REAL',
                 'CAL-CS_TDEP_PCAL_LINE2_REF_A_UIM_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE2_CORRECTION_REAL',
                 'CAL-CS_TDEP_PCAL_LINE2_CORRECTION_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_C_NOCAVPOLE_REAL',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_C_NOCAVPOLE_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_D_REAL',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_D_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_A_TST_REAL',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_A_TST_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_A_PUM_REAL',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_A_PUM_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_A_UIM_REAL',
                 'CAL-CS_TDEP_PCAL_LINE4_REF_A_UIM_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE4_CORRECTION_REAL',
                 'CAL-CS_TDEP_PCAL_LINE4_CORRECTION_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_C_NOCAVPOLE_REAL',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_C_NOCAVPOLE_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_D_REAL',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_D_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_A_TST_REAL',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_A_TST_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_A_PUM_REAL',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_A_PUM_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_A_UIM_REAL',
                 'CAL-CS_TDEP_PCAL_LINE1_REF_A_UIM_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE3_CORRECTION_REAL',
                 'CAL-CS_TDEP_PCAL_LINE3_CORRECTION_IMAG',
                 'CAL-CS_TDEP_PCAL_LINE1_PCAL_DEMOD_PHASE',
                 'CAL-CS_TDEP_PCAL_LINE2_PCAL_DEMOD_PHASE',
                 'CAL-CS_TDEP_PCAL_LINE3_PCAL_DEMOD_PHASE',
                 'CAL-CS_TDEP_PCAL_LINE4_PCAL_DEMOD_PHASE',
                 'CAL-CS_TDEP_SUS_LINE1_SUS_DEMOD_PHASE',
                 'CAL-CS_TDEP_SUS_LINE2_SUS_DEMOD_PHASE',
                 'CAL-CS_TDEP_SUS_LINE3_SUS_DEMOD_PHASE'))

            save_txt_epics = np.column_stack(
                (save_txt_epics_channel_name,
                 save_txt_epcis_value))

            utc_now = from_gps(tconvert())

            np.savetxt(
                f'epicsrecords_{self.name}'
                f'_model{self.model_date}'
                f'_create{utc_now.year}{utc_now.month}{utc_now.day}.txt',
                save_txt_epics,
                fmt='%s %.6e')

        # Output - see T1700106
        return {
            'CAL-CS_TDEP_PCAL_LINE1_CORRECTION': PCAL_LINE1_CORRECTION,
            'CAL-CS_TDEP_SUS_LINE3_REF_INVA_TST_RESPRATIO': SUS_LINE3_REF_INVA_TST_RESPRATIO,
            'CAL-CS_TDEP_SUS_LINE2_REF_INVA_PUM_RESPRATIO': SUS_LINE2_REF_INVA_PUM_RESPRATIO,
            'CAL-CS_TDEP_SUS_LINE1_REF_INVA_UIM_RESPRATIO': SUS_LINE1_REF_INVA_UIM_RESPRATIO,
            'CAL-CS_TDEP_PCAL_LINE2_REF_C_NOCAVPOLE': PCAL_LINE2_REF_C_NOCAVPOLE,
            'CAL-CS_TDEP_PCAL_LINE2_REF_D': PCAL_LINE2_REF_D,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_TST': PCAL_LINE2_REF_A_TST,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_PUM': PCAL_LINE2_REF_A_PUM,
            'CAL-CS_TDEP_PCAL_LINE2_REF_A_UIM': PCAL_LINE2_REF_A_UIM,
            'CAL-CS_TDEP_PCAL_LINE2_CORRECTION': PCAL_LINE2_CORRECTION,
            'CAL-CS_TDEP_PCAL_LINE4_REF_C_NOCAVPOLE': PCAL_LINE4_REF_C_NOCAVPOLE,
            'CAL-CS_TDEP_PCAL_LINE4_REF_D': PCAL_LINE4_REF_D,
            'CAL-CS_TDEP_PCAL_LINE4_REF_A_TST': PCAL_LINE4_REF_A_TST,
            'CAL-CS_TDEP_PCAL_LINE4_REF_A_PUM': PCAL_LINE4_REF_A_PUM,
            'CAL-CS_TDEP_PCAL_LINE4_REF_A_UIM': PCAL_LINE4_REF_A_UIM,
            'CAL-CS_TDEP_PCAL_LINE4_CORRECTION': PCAL_LINE4_CORRECTION,
            'CAL-CS_TDEP_PCAL_LINE1_REF_C_NOCAVPOLE': PCAL_LINE1_REF_C_NOCAVPOLE,
            'CAL-CS_TDEP_PCAL_LINE1_REF_D': PCAL_LINE1_REF_D,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_TST': PCAL_LINE1_REF_A_TST,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_PUM': PCAL_LINE1_REF_A_PUM,
            'CAL-CS_TDEP_PCAL_LINE1_REF_A_UIM': PCAL_LINE1_REF_A_UIM,
            'CAL-CS_TDEP_PCAL_LINE3_CORRECTION': PCAL_LINE3_CORRECTION,
            'CAL-CS_TDEP_PCAL_LINE1_PCAL_DEMOD_PHASE': PCAL_LINE1_PCAL_DEMOD_PHASE,
            'CAL-CS_TDEP_PCAL_LINE2_PCAL_DEMOD_PHASE': PCAL_LINE2_PCAL_DEMOD_PHASE,
            'CAL-CS_TDEP_PCAL_LINE3_PCAL_DEMOD_PHASE': PCAL_LINE3_PCAL_DEMOD_PHASE,
            'CAL-CS_TDEP_PCAL_LINE4_PCAL_DEMOD_PHASE': PCAL_LINE4_PCAL_DEMOD_PHASE,
            'CAL-CS_TDEP_SUS_LINE1_SUS_DEMOD_PHASE': SUS_LINE1_SUS_DEMOD_PHASE,
            'CAL-CS_TDEP_SUS_LINE2_SUS_DEMOD_PHASE': SUS_LINE2_SUS_DEMOD_PHASE,
            'CAL-CS_TDEP_SUS_LINE3_SUS_DEMOD_PHASE': SUS_LINE3_SUS_DEMOD_PHASE,
        }

    def plot(self, plot_selection='all', freq_min=0.1, freq_max=5000, filename=None, ifo='',
             label=None, style=None, ugf_start=10, ugf_end=1000, show=None, **kwargs):
        """
        Make DARM critique plots.
        This method produces critique models for 1 or 2 models.

        PARAMETERS
        ----------
        args: The first argumets should be 1 or 2 darm models.
                  - darm model object (DARMModel)

        The plots produced should be close to the final critique plot, but if
        one needs something special there are a few options to tw
        Options for the graphs: - freq_min (float) start frequency
                                - freq_max (float) end frequency
                                - **kwargs, you can also use all the available
                                            matplotlib keywords, but have to be common
                                            for all graphs.
                                - filename (string)
                                  if given, ALL generated graphs will be saved in one pdf.
                                - ifo (string)
                                  if given with a model to plot, it will appear in the
                                  graph titles.
                                - ugf_start, ugf_end (float)
                                  frequency range used for the search of the UGF
                                - greed (bool)
                                - plot_selection (string)
                                  'all':
                                  'optical':
                                  'actuation':
                                  'clg':
                                  'olg':
                                  'digital':
        RETURNS
        -------
        Plots critique plots of the given darm model, and saves them in one pdf file.
        """

        critique(self, freq_min=freq_min, freq_max=freq_max, filename=filename, label=label,
                 plot_selection=plot_selection, ifo=ifo, show=show,
                 ugf_start=ugf_start, ugf_end=ugf_end, **kwargs)



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy import signal, io, constants
from collections import namedtuple

from .utils import (serielZPK, digital_delay_filter, load_foton_export_tf,
                    compute_digital_filter_response)
from .digital import iopdownsamplingfilters
from .model import Model


class SensingModel(Model):
    """
    A sensing function model object

    This is a class to set up the model for the sensing function from a
    configuration file with all the information about where the data is stored

    Parameters
    ----------

    Returns
    -------


    """

    def __init__(self, config):
        super().__init__(config, measurement='sensing')

        # TODO need some way of validating the configuration file is
        # appropriate

    @staticmethod
    def optical_response(pole_frequency, detune_spring_frequency,
                         detune_spring_Q, pro_spring=False):
        """
        Compute the coupled cavity response approximated by a single cavity
        pole and a detuned spring frequency and Q

        Parameters
        ----------
        pole_frequency : `float`
            Coupled cavity pole frequency in Hz
        detune_spring_frequency : `float`
            Detuning spring frequency in Hz
        detune_spring_Q : `float`
            Unitless Q
        pro_spring : `boolean`, optional
            logic flag to indicate pro-spring

        Returns
        -------
        optical_response : `LTI object`
            coupled cavity optical response
        """

        cavity_pole = signal.ZerosPolesGain(
            [], -2.0*np.pi*pole_frequency, 2.0*np.pi*pole_frequency)
        if not pro_spring:
            # Anti-spring: f^2 / (f^2 + fs^2 - 1j * f * fs / Q)
            # see alog LHO 48057
            detuning = signal.TransferFunction(
                [1, 0, 0],
                [1, 2.0*np.pi*detune_spring_frequency/detune_spring_Q,
                 -(2.0*np.pi*detune_spring_frequency)**2])
            detuning = detuning.to_zpk()
        else:
            # Pro-spring: f^2 / (f^2 - fs^2 + 1j * f * fs / Q)
            # see alog LHO 48057
            detuning = signal.TransferFunction(
                [1, 0, 0],
                [1, -2.0*np.pi*detune_spring_frequency/detune_spring_Q,
                 (2.0*np.pi*detune_spring_frequency)**2])
            detuning = detuning.to_zpk()
        coupled_cavity = serielZPK(cavity_pole, detuning)
        return coupled_cavity

    def omc_analog_dcpd_readout_response(self, name, frequencies):
        """
        The transfer function of the analog OMC DCPD readout electronics

        Parameters
        ----------
        name : str
            a string (e.g., 'A' or 'A_A') indicating the OMC path name to be
            evaluated. This string should be listed in
            SensingModel.omc_path_names variable
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        response : `complex128`, array-like
            transfer function response of the analog electronics in path `name`
        """

        # Make sure that name is listed uniquely in self.omc_path_names
        idx = self.omc_path_names.index(name)
        assert len(set(self.omc_path_names)) == len(self.omc_path_names)

        # First the transimpedence amplifier is computed
        if (hasattr(self, 'omc_meas_z_trans_amplifier') and
                hasattr(self, 'omc_meas_p_trans_amplifier')):
            if idx < len(self.omc_meas_z_trans_amplifier):
                meas_z_trans_amp = np.asarray(
                    self.omc_meas_z_trans_amplifier[idx])
            else:
                meas_z_trans_amp = np.asarray([])
            meas_p_trans_amp = np.asarray(
                self.omc_meas_p_trans_amplifier[idx])
        elif hasattr(self, 'omc_meas_p_trans_amplifier'):
            meas_z_trans_amp = np.asarray([])
            meas_p_trans_amp = np.asarray(
                self.omc_meas_p_trans_amplifier[idx])
        assert len(meas_z_trans_amp) <= len(meas_p_trans_amp)
        if len(meas_p_trans_amp) > 0:
            omc_trans_amplifier_response = signal.freqresp(
                signal.ZerosPolesGain(-2*np.pi*meas_z_trans_amp,
                                      -2*np.pi*meas_p_trans_amp,
                                      np.prod(2*np.pi*meas_p_trans_amp) /
                                      np.prod(2*np.pi*meas_z_trans_amp)),
                2*np.pi*frequencies)[1]
        else:
            omc_trans_amplifier_response = np.ones(len(frequencies),
                                                   dtype='complex128')

        # Whitening filter
        # The zeros and poles need to have the name appended to them
        whitening_name = getattr(self, 'whitening_mode_names')[idx]
        if (hasattr(self, f'omc_meas_z_whitening_{whitening_name}') and
                hasattr(self, f'omc_meas_p_whitening_{whitening_name}')):
            zeros = getattr(self, f'omc_meas_z_whitening_{whitening_name}')
            if idx < len(zeros):
                meas_z_whitening = np.asarray(zeros[idx])
            else:
                meas_z_whitening = np.asarray([])
            poles = getattr(self, f'omc_meas_p_whitening_{whitening_name}')
            meas_p_whitening = np.asarray(poles[idx])
        elif getattr(self, f'omc_meas_p_whitening_{whitening_name}'):
            meas_z_whitening = np.asarray([])
            poles = getattr(self, f'omc_meas_p_whitening_{whitening_name}')
            meas_p_whitening = np.asarray(poles[idx])
        assert len(meas_z_whitening) <= len(meas_p_whitening)
        if len(meas_p_whitening) > 0:
            omc_whitening_response = signal.freqresp(
                signal.ZerosPolesGain(-2*np.pi*meas_z_whitening,
                                      -2*np.pi*meas_p_whitening,
                                      np.prod(2*np.pi*meas_p_whitening) /
                                      np.prod(2*np.pi*meas_z_whitening)),
                2*np.pi*frequencies)[1]
        else:
            omc_whitening_response = np.ones(len(frequencies),
                                             dtype='complex128')

        # super high frequency poles look like a delay (see G2200551)
        super_high_frequency_poles_apparent_delay = np.exp(
            -2.0*np.pi*1j *
            self.super_high_frequency_poles_apparent_delay[idx] *
            frequencies)

        # Normalized analog anti-aliasing frequency response
        analog_aa_filter_response = \
            self.analog_aa_or_ai_filter_response(frequencies, idx)
        norm = np.abs(self.analog_aa_or_ai_filter_response(1e-2,
                                                           idx))
        analog_aa_filter_response_hi_f = analog_aa_filter_response / norm

        # Return a the combined transfer function response
        return (omc_trans_amplifier_response * omc_whitening_response *
                analog_aa_filter_response_hi_f *
                super_high_frequency_poles_apparent_delay)

    def omc_digital_compensation_response(self, name, frequencies):
        """
        The transfer function of the OMC digital compensation filters

        Parameters
        ----------
        name : str
            a string (e.g., 'A' or 'A_A') indicating the OMC path name to be
            evaluated. This string should be listed in
            SensingModel.omc_path_names variable
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        compensation_response : `complex128`, array-like
            transfer function response of the digital compensation filters in
            the path labeled by `name`
        """

        idx = self.omc_path_names.index(name)
        assert len(set(self.omc_path_names)) == len(self.omc_path_names)

        # Always return a compensation of 1 for all frequencies if nothing is
        # provided in the configuration
        compensation_response = np.ones(len(frequencies), dtype='complex128')

        # If filter file provided and filter bank provided with at least one
        # element, then we'll get a transfer function
        if (hasattr(self, 'omc_compensation_filter_file') and
                self.omc_compensation_filter_file != '' and
                len(self.omc_compensation_filter_bank) > idx and
                self.omc_compensation_filter_bank[idx] != '' and
                len(self.omc_compensation_filter_modules_in_use) > idx and
                self.omc_compensation_filter_modules_in_use[idx] != '' and
                len(self.omc_compensation_filter_gain) > idx and
                self.omc_compensation_filter_gain[idx] != ''):
            tf, pfilt = compute_digital_filter_response(
                self.dpath(self.omc_compensation_filter_file),
                self.omc_compensation_filter_bank[idx],
                self.omc_compensation_filter_modules_in_use[idx],
                self.omc_compensation_filter_gain[idx],
                frequencies)
            compensation_response *= tf
        elif (hasattr(self, 'omc_compensation_filter_file') and
                self.omc_compensation_filter_file != '' and
                len(self.omc_compensation_filter_bank) > idx and
                self.omc_compensation_filter_bank[idx] != '' and
                len(self.omc_compensation_filter_gain) > idx and
                self.omc_compensation_filter_gain[idx] != ''):
            compensation_response *= self.omc_compensation_filter_gain[idx]

        return compensation_response

    def omc_path_response(self, name, frequencies):
        """
        Compute the frequency response for a single OMC DCPD readout as shown
        in G1501518-v21 (the bracketed term directly below "O4 Calibration").
        This includes the gain ratio (= path idx / ref path, usually path A)
        and the balance matrix coefficient for path idx.

        The reference path (usually path A) should have gain ratio = 1.

        Parameters
        ----------
        name : str
            a string (e.g., 'A' or 'A_A') indicating the OMC path name to be
            evaluated. This string should be listed in
            SensingModel.omc_path_names variable
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        path_response : `complex128`, array-like
            transfer function response of the path labeled by `idx`
        """

        idx = self.omc_path_names.index(name)
        assert len(set(self.omc_path_names)) == len(self.omc_path_names)

        # analog OMC readout electronics
        analog_readout_response = self.omc_analog_dcpd_readout_response(
            name, frequencies)

        # ADC gain
        adc_gain = self.adc_gain[idx]

        # front end digital compensation of analog OMC electronics
        omc_fe_tf = self.omc_digital_compensation_response(name, frequencies)

        # digital anti-aliasing filter
        digital_aa_filter_response = \
            self.digital_aa_or_ai_filter_response(frequencies)

        # Gain ratio (note this does not include the balance matrix)
        gain_ratio = self.gain_ratio[idx]

        # balance matrix
        matrix_value = self.balance_matrix[idx]

        # The complete path response is:
        # OMC analog readout response (trans. amp, whitening, analog AA)
        # OMC digital compensation, if any
        # OMC digital AA filtering
        # Gain ratio
        # Matrix element
        path_response = analog_readout_response * adc_gain * omc_fe_tf * \
            digital_aa_filter_response * gain_ratio * matrix_value

        return path_response

    def omc_combine_path_responses(self, frequencies):
        """
        Compute the frequency response from the OMC DCPD electronics as shown
        in G1501518-v21 (the bracketed term directly below "O4 Calibration")

        Note however, that this is the WEIGHTED MEAN! Meaning that for N paths,
        we would sum the paths and divide by N if they were perfectly balanced

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        combined_paths : `complex128`, array-like
            transfer function response
        """

        # Initialize the total weight to be 0
        sum_of_weights = 0
        response = np.zeros(len(frequencies), dtype='complex128')

        # For each OMC path, add the path response to the overall response
        # and add the gain ratio value x balance matrix value for that path
        # to the sum_of_weights variable
        for n in range(len(self.omc_path_names)):
            response += self.omc_path_response(self.omc_path_names[n],
                                               frequencies)
            sum_of_weights += \
                (self.gain_ratio[n] * self.balance_matrix[n])

        # Divide the response by the sum_of_weights to calculate the weighted
        # mean OMC path response
        response /= sum_of_weights

        return response

    def light_travel_time_delay_response(self, frequencies):
        """
        Compute the frequency response from the light travel time delay

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response
        """

        return np.exp(-2.0*np.pi*1j *
                      np.mean(np.array([self.x_arm_length,
                                        self.y_arm_length])) /
                      constants.c * frequencies)

    def single_pole_approximation_delay_correction_response(self, frequencies):
        """
        Compute the frequency response from the time delay correction to the
        single pole approximation

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response
        """

        return np.exp(-2.0*np.pi*1j *
                      self.single_pole_approximation_delay_correction *
                      frequencies)

    def known_sensing_terms_for_measurement(self, frequencies):
        """
        Compute the known terms in the sensing measurement that we account
        for by dividing out. See G1501518-v21

        Note that we are now allocating all gain on the sensing path to
        the overall optical gain (previously a 0.99 gain on AA was not
        allocated to the optical gain)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the known sensing terms

        Returns
        -------
        sensing_correction : `complex128`, array-like
            correction terms
        """

        light_travel_time_delay = \
            self.light_travel_time_delay_response(frequencies)
        single_pole_approximation_delay_correction = \
            self.single_pole_approximation_delay_correction_response(
                frequencies)
        omc_paths_response = self.omc_combine_path_responses(frequencies)

        sensing_correction = \
            self.sensing_sign * \
            light_travel_time_delay * \
            single_pole_approximation_delay_correction * \
            omc_paths_response

        return sensing_correction

    def compute_sensing(self, frequencies):
        """
        Compute the entire sensing function transfer function (see G1501518)

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the sensing function
        """

        # coupled cavity response without the optical gain
        coupled_cavity = self.optical_response(
            self.coupled_cavity_pole_frequency, self.detuned_spring_frequency,
            self.detuned_spring_q, pro_spring=self.is_pro_spring)
        coupled_cavity_filter_response = \
            signal.freqresp(coupled_cavity, 2.0*np.pi*frequencies)[1]

        # By design there are no digital delays to the OMC user model
        # see G1601472

        # light travel time delay
        light_travel_time_delay = \
            self.light_travel_time_delay_response(frequencies)

        # single pole approximation needs a correction for too much delay
        # see G1501316
        single_pole_approximation_delay_correction = \
            self.single_pole_approximation_delay_correction_response(
                frequencies)

        # OMC paths response
        omc_paths_response = self.omc_combine_path_responses(frequencies)

        # All together:
        # sensing sign
        # overall optical gain
        # coupled-cavity filter response
        # light travel time delay
        # single pole approximation delay correction
        # OMC DCPD electronics paths
        tf = self.sensing_sign * \
            self.coupled_cavity_optical_gain * \
            coupled_cavity_filter_response * \
            light_travel_time_delay * \
            single_pole_approximation_delay_correction * \
            omc_paths_response

        return tf

    def sensing_residual(self, frequencies):
        """
        Compute the residual sensing function

        This is everything in the sensing function except the optical gain
        and the optical response

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute displacement to CAL-CW residual

        Returns
        -------
        C_res : `complex128`, array-like
            transfer function response
        """

        coupled_cavity = self.optical_response(
            self.coupled_cavity_pole_frequency,
            self.detuned_spring_frequency,
            self.detuned_spring_q,
            pro_spring=self.is_pro_spring)
        C_res = (self.compute_sensing(frequencies) / (
            self.coupled_cavity_optical_gain *
            signal.freqresp(coupled_cavity, 2.0*np.pi*frequencies)[1]))

        return C_res

    def displacement_to_calcs_residual(self, frequencies):
        """
        Compute the ETM displacement to CAL-CS residual

        This will be C * 1 16k clock cycle delay * foton inverse sensing filter

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute displacement to CAL-CW residual

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response
        """

        # Get the sensing model
        C_pydarm = self.compute_sensing(frequencies)

        # Need the coupled cavity model
        coupled_cavity = self.optical_response(
            self.coupled_cavity_pole_frequency, self.detuned_spring_frequency,
            self.detuned_spring_q, pro_spring=self.is_pro_spring)

        # Need one 16k clock cycle delay going from OMC to CAL-CS
        one_clock_cycle_filter = digital_delay_filter(1, 2**14)
        one_clock_cycle_filter_response = \
            signal.dfreqresp(one_clock_cycle_filter,
                             2.0*np.pi*frequencies/2**14)[1]

        # load inverse sensing data from foton file export (1/SRC_D2N)
        if hasattr(self, 'gds_c_foton_invsensing_tf') and \
                self.gds_c_foton_invsensing_tf != '':
            [foton_export_freq, foton_export_tf] = \
                load_foton_export_tf(
                    self.dpath(self.gds_c_foton_invsensing_tf))
        else:
            # This is the fake if no foton export file is provided
            foton_export_freq = frequencies
            foton_export_tf = 1.0 / \
                signal.freqresp(
                    coupled_cavity,
                    2.0*np.pi*foton_export_freq)[1]

        # interpolate the foton filter data to the frequencies used
        foton_inv_sensing_interp = np.interp(frequencies,
                                             foton_export_freq,
                                             foton_export_tf)

        return C_pydarm * \
            one_clock_cycle_filter_response * \
            foton_inv_sensing_interp

    def old_school_senspars(self, frequencies):
        """
        The old version of pyDARM stored intermediate data products in a
        namedtuple for use when making other computations. This is not normally
        used for this version of pyDARM since there are now function calls
        but for backwards compatibility with other scripts, there is a request
        that this be available for use. NOTE: this currently does not contain
        everything, like the pcalCorrection or OMC delay

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute intermediate data products
            responses

        Returns
        -------
        sensingProducts : namedtuple
            namedtuple of all the intermediate data products
        """
        sensingProducts = \
            namedtuple('sensingProducts',
                       ['coupledCavity', 'AAanalog', 'AAanalog_delay',
                        'AAdigital', 'uncompensatedOMCDCPD', 'omcDCPDAtf',
                        'omcDCPDBtf', 'omcDCPDtf', 'delay', 'sensIOPdelay',
                        'omcDelay', 'omcDelay_freqresp',
                        'totalSensingDigital', 'totalSensingDigital_freqresp',
                        'totalSensingAnalog_freqresp',
                        'totalSensingAnalogNoCav_freqresp',
                        'singlePoleApproxDelayCorrection', 'pcaldewhitening',
                        'pcalCorrection', 'C_NoCav',
                        'calcsResid2gdsInvSensOut', 'displ2calcsResid',
                        'IIRwarp_interp'])

        coupled_cavity = self.optical_response(
            self.coupled_cavity_pole_frequency, self.detuned_spring_frequency,
            self.detuned_spring_q, pro_spring=self.is_pro_spring)
        mat = io.loadmat(self.dpath(self.analog_anti_aliasing_file))
        analog_aa_zpk = signal.ZerosPolesGain(mat['AAzeros'][:, 0],
                                              mat['AApoles'][:, 0],
                                              mat['AAgain'][0, 0])
        analog_aa_delay = np.exp(
            -2.0*np.pi*mat['AAdelay'][0, 0]*1j*frequencies)
        digital_aa_filter = iopdownsamplingfilters(
            self.anti_aliasing_rate_string, self.anti_aliasing_method,
            rcg_ver=3)
        digital_aa_filter = digital_aa_filter.to_zpk()
        uncompensated_omc_poles_trans_whitening_mean = \
            0.5*(self.uncompensated_omc_poles_trans_whitening_a +
                 self.uncompensated_omc_poles_trans_whitening_b)
        combined_uncompensated_poles = np.append(
              self.uncompensated_omc_poles_trans_amplifier,
              uncompensated_omc_poles_trans_whitening_mean)
        uncompensated_omc_dcpd = \
            signal.ZerosPolesGain(
                [], -2.0*np.pi*combined_uncompensated_poles,
                np.prod(2.0*np.pi*combined_uncompensated_poles))
        light_travel_time_delay = \
            self.light_travel_time_delay_response(frequencies)
        digital_aa_filter_response = signal.dfreqresp(
            digital_aa_filter, 2.0*np.pi*frequencies/2**16)[1]
        single_pole_approximation_delay_correction = \
            self.single_pole_approximation_delay_correction_response(
                frequencies)
        sensing_analog_nocav = self.coupled_cavity_optical_gain * \
            self.sensing_sign * \
            light_travel_time_delay * \
            single_pole_approximation_delay_correction * \
            signal.freqresp(uncompensated_omc_dcpd, 2*np.pi*frequencies)[1] * \
            signal.freqresp(analog_aa_zpk, 2*np.pi*frequencies)[1] * \
            analog_aa_delay
        sensing_analog = sensing_analog_nocav * \
            signal.freqresp(coupled_cavity, 2*np.pi*frequencies)[1]
        sensing_nocav = self.compute_sensing(frequencies) / \
            signal.freqresp(coupled_cavity, 2*np.pi*frequencies)[1]
        if self.gds_c_foton_invsensing_tf is not None:
            foton_inv_sensing_data = np.loadtxt(
                self.dpath(self.gds_c_foton_invsensing_tf))
            foton_inv_sensing_frequencies = foton_inv_sensing_data[:, 0]
            foton_inv_sensing_response = foton_inv_sensing_data[:, 1] + \
                1j*foton_inv_sensing_data[:, 2]
        else:
            foton_inv_sensing_frequencies = frequencies
            foton_inv_sensing_response = np.ones(len(frequencies))
        coupled_cavity_response = \
            signal.freqresp(coupled_cavity,
                            2.0*np.pi*foton_inv_sensing_frequencies)[1]
        foton_inv_sensing_warp = (1.0/coupled_cavity_response) / \
            foton_inv_sensing_response

        # interpolate the foton filter data to the frequencies used
        foton_inv_sensing_interp = np.interp(frequencies,
                                             foton_inv_sensing_frequencies,
                                             foton_inv_sensing_warp)

        return sensingProducts(coupled_cavity, analog_aa_zpk, analog_aa_delay,
                               digital_aa_filter, uncompensated_omc_dcpd,
                               [], [], [], light_travel_time_delay, [],
                               [], [], digital_aa_filter,
                               digital_aa_filter_response, sensing_analog,
                               sensing_analog_nocav,
                               single_pole_approximation_delay_correction,
                               [], [], sensing_nocav,
                               self.gds_sensing_correction(frequencies),
                               self.displacement_to_calcs_residual(
                                   frequencies),
                               foton_inv_sensing_interp)
